// core functions to generate initial states and evolve them

// get the next evolution of a 1D CA initial state
// buggy BUT produces interesting results
// function evolve1d(state, rules) {
//   const sl = state.length - 1;
//   return [
//     rules[[state[sl - 1], state[sl], state[sl + 1]].join("")],
//     ...state.map((_,x) => (rules[[state[x - 1], state[x], state[x + 1]].join("")])).slice(0, sl),
//     rules[[state[0], state[1], state[2]].join("")]
//   ]
// }

// get the next evolution of a 1D CA initial state
function evolve1d(state, rules) {
  const sl = state.length - 1;
  const edge1 = [state[sl - 2], state[sl - 1], state[sl]].join("");
  const edge2 = [state[0], state[1], state[2]].join("");
  // normal case (3 neighbor cells)
  const center = state
    .map((_, x) => rules[[state[x - 1], state[x], state[x + 1]].join("")])
    .slice(1, sl);
  return [rules[edge1], ...center, rules[edge2]];
}

// create a 2D board from a 1D CA initial state
// function createBoard(state, rules, height) {
//   function createBoardAcc(s, h, acc) {
//     if (h === 0) return acc;
//     const newState = evolve1d(s, rules);
//     const newAcc = acc.concat([s]);
//     return createBoardAcc(newState, h - 1, newAcc);
//   }
//   return createBoardAcc(state, height, []);
// }

// performance "choke point" in full imperative
function createBoard(state, rules, max) {
  var board = [];
  let prevState = [];
  for (let i = 0; i < max; i++) {
    let nextState = [];
    // use the passed initial step during first iteration
    if (i == 0) {
      nextState = evolve1d(state, rules);
    } else {
      nextState = evolve1d(prevState, rules);
    }
    board = board.concat([nextState]);
    prevState = nextState;
  }
  return board;
}

// Sums the value of a cell's neighbors
function getNeighborsSum(cells) {
  return cells.reduce((cell, acc) => cell + acc, 0);
}

// Get the next evolution of a cell according to
// Conway's game of life rules
function conwayRules(cell, neighbors) {
  // loneliness rule
  if (cell === 1 && neighbors < 2) return 0;
  // overpopulation rule
  if (cell === 1 && neighbors > 3) return 0;
  // born when three live neighbors rule
  if (cell === 0 && neighbors === 3) return 1;
  // the cell remains the same if none apply
  return cell;
}

// Get the next evolution of a cell according to
// Conway's game of life rules
function servietteRules(cell, neighbors) {
  // loneliness rule
  if (cell === 0 && [2, 3, 4].find((x) => x == neighbors)) return 1;
  // the cell remains the same if none apply
  return 0;
}

// variation of the game of life where a
// cell comes to live if 6 neigbor cells are alive
function highLifeRules(cell, neighbors) {
  // loneliness rule
  if (cell === 1 && neighbors < 2) return 0;
  // overpopulation rule
  if (cell === 1 && neighbors > 3) return 0;
  // born when three live neighbors rule
  if (cell === 0 && neighbors === 2) return 1;
  // highlife rules
  if ((cell === 0 && neighbors === 3) || neighbors === 6) return 1;
  // the cell remains the same if none apply
  return cell;
}

// variation on the game of life's rules,
// where the "three live neighbors" rule is ignored
function threebornRules(cell, neighbors) {
  // loneliness rule
  if (cell === 1 && neighbors < 2) return 0;
  // overpopulation rule
  if (cell === 1 && neighbors > 3) return 0;
  // born when three live neighbors rule
  return cell;
}

// variation on the game of life's rules,
// where the loneliness rule is ignored
function lonelinessRules(cell, neighbors) {
  // overpopulation rule
  if (cell === 1 && neighbors > 3) return 0;
  // born when three live neighbors rule
  if (cell === 0 && neighbors === 3) return 1;
  // the cell remains the same if none apply
  return cell;
}

// variation on the game of life's rules,
// where the overpopulation rule is ignored
function overpopulationRules(cell, neighbors) {
  // loneliness rule
  if (cell === 1 && neighbors < 2) return 0;
  // born when three live neighbors rule
  if (cell === 0 && neighbors === 3) return 1;
  // the cell remains the same if none apply
  return cell;
}

// get the next evolution of a 2D CA initial state
// Rules : Moore neighborhood
function evolve2d(board, rulesFn) {
  const bh = board.length - 1;
  const bw = board[0].length - 1;
  return board.map((row, y) => {
    // handle edges
    const prow = y - 1 < 0 ? board[bh] : board[y - 1];
    const nrow = y + 1 > bh ? board[0] : board[y + 1];
    return row.map((cell, x) => {
      // handle edges too
      const pcell = x - 1 < 0 ? bw : x - 1;
      const ncell = x + 1 > bw ? 0 : x + 1;
      // the current cell is not included in the result
      const neighbors = [
        prow[pcell],
        prow[x],
        prow[ncell],
        row[pcell],
        row[ncell],
        nrow[pcell],
        nrow[x],
        nrow[ncell],
      ];
      const sum = getNeighborsSum(neighbors);
      return rulesFn(cell, sum);
    });
  });
}

function getDrawingValues(state, acc, cell) {
  const d = cell.dimension;
  return Object.keys(state).map((key) => {
    const fillStyle = (() => {
      if (state[key] === "1") return cell.liveColor;
      return cell.deadColor;
    })();

    return {
      move: [key * d, acc * d],
      fill: [key * d, acc * d, d, d],
      fillStyle,
    };
  });
}

// Populates the first state with a single living cell in the center
function create1dStateOneCell(width) {
  return [...Array(width)].map((cell, index) => {
    if (index === Math.floor(width / 2)) return 1;
    return 0;
  });
}

// Populates the first state of a 1D CA with cells returned
// by initFn
function create1dState(width, initFn, args) {
  return [...Array(width)].map(() => initFn(...args));
}

// Populates the first state of a 2D CA with cells returned
// by initFn
function create2dState(width, height, initFn, args) {
  return [...Array(height)].map(() =>
    [...Array(width)].map(() => initFn(...args))
  );
}

export {
  getDrawingValues,
  create1dState,
  create2dState,
  createBoard,
  create1dStateOneCell,
  conwayRules,
  overpopulationRules,
  lonelinessRules,
  threebornRules,
  highLifeRules,
  servietteRules,
  evolve1d,
  evolve2d,
};
